@@ -1,11 +1,11 @@
 # -*- test-case-name: twisted.conch.test.test_cftp -*-
-# Copyright (c) 2001-2009 Twisted Matrix Laboratories.
+# Copyright (c) Twisted Matrix Laboratories.
 # See LICENSE for details.
 
 """
 Implementation module for the I{cftp} command.
 """
-
+from __future__ import division, print_function
 import os, sys, getpass, struct, tty, fcntl, stat
 import fnmatch, pwd, glob
 
@@ -30,12 +30,12 @@ class ClientOptions(options.ConchOptions):
                     ['batchfile', 'b', None, 'File to read commands from, or \'-\' for stdin.'],
                     ['requests', 'R', 5, 'Number of requests to make before waiting for a reply.'],
                     ['subsystem', 's', 'sftp', 'Subsystem/server program to connect to.']]
-    zsh_altArgDescr = {"buffersize":"Size of send/receive buffer (default: 32768)"}
-    #zsh_multiUse = ["foo", "bar"]
-    #zsh_mutuallyExclusive = [("foo", "bar"), ("bar", "baz")]
-    #zsh_actions = {"foo":'_files -g "*.foo"', "bar":"(one two three)"}
-    #zsh_actionDescr = {"logfile":"log file name", "random":"random seed"}
-    zsh_extras = ['2::localfile:{if [[ $words[1] == *:* ]]; then; _files; fi}']
+
+    compData = usage.Completions(
+        descriptions={
+            "buffersize": "Size of send/receive buffer (default: 32768)"},
+        extraActions=[usage.CompleteUserAtHost(),
+                      usage.CompleteFiles(descr="local file")])
 
     def parseArgs(self, host, localPath=None):
         self['remotePath'] = ''
@@ -57,8 +57,8 @@ def run():
     options = ClientOptions()
     try:
         options.parseOptions(args)
-    except usage.UsageError, u:
-        print 'ERROR: %s' % u
+    except usage.UsageError as u:
+        print('ERROR: %s' % u)
         sys.exit(1)
     if options['log']:
         realout = sys.stdout
@@ -105,7 +105,7 @@ def _ebExit(f):
         s = f.value.value
     else:
         s = str(f)
-    print s
+    print(s)
     #exitStatus = "conch: exiting with error %s" % f
     try:
         reactor.stop()
@@ -400,63 +400,148 @@ class StdioClient(basic.LineReceiver):
             self.transport.write('\n')
         return "Transferred %s to %s" % (rf.name, lf.name)
 
+
     def cmd_PUT(self, rest):
+        """
+        Do an upload request for a single local file or a globing expression.
+
+        @param rest: Requested command line for the PUT command.
+        @type rest: C{str}
+
+        @return: A deferred which fires with C{None} when transfer is done.
+        @rtype: L{defer.Deferred}
+        """
         local, rest = self._getFilename(rest)
-        if '*' in local or '?' in local: # wildcard
+
+        # FIXME: https://twistedmatrix.com/trac/ticket/7241
+        # Use a better check for globbing expression.
+        if '*' in local or '?' in local:
             if rest:
                 remote, rest = self._getFilename(rest)
-                path = os.path.join(self.currentDirectory, remote)
-                d = self.client.getAttrs(path)
-                d.addCallback(self._cbPutTargetAttrs, remote, local)
-                return d
+                remote = os.path.join(self.currentDirectory, remote)
             else:
                 remote = ''
-                files = glob.glob(local)
-                return self._cbPutMultipleNext(None, files, remote)
-        if rest:
-            remote, rest = self._getFilename(rest)
+
+            files = glob.glob(local)
+            return self._putMultipleFiles(files, remote)
+
         else:
-            remote = os.path.split(local)[1]
-        lf = file(local, 'r')
-        path = os.path.join(self.currentDirectory, remote)
-        flags = filetransfer.FXF_WRITE|filetransfer.FXF_CREAT|filetransfer.FXF_TRUNC
-        d = self.client.openFile(path, flags, {})
-        d.addCallback(self._cbPutOpenFile, lf)
-        d.addErrback(self._ebCloseLf, lf)
-        return d
+            if rest:
+                remote, rest = self._getFilename(rest)
+            else:
+                remote = os.path.split(local)[1]
+            return self._putSingleFile(local, remote)
 
-    def _cbPutTargetAttrs(self, attrs, path, local):
-        if not stat.S_ISDIR(attrs['permissions']):
-            return "Wildcard put with non-directory target."
-        return self._cbPutMultipleNext(None, files, path)
 
-    def _cbPutMultipleNext(self, res, files, path):
-        if isinstance(res, failure.Failure):
-            self._printFailure(res)
-        elif res:
-            self.transport.write(res)
-            if not res.endswith('\n'):
+    def _putSingleFile(self, local, remote):
+        """
+        Perform an upload for a single file.
+
+        @param local: Path to local file.
+        @type local: C{str}.
+
+        @param remote: Remote path for the request relative to current working
+            directory.
+        @type remote: C{str}
+
+        @return: A deferred which fires when transfer is done.
+        """
+        return self._cbPutMultipleNext(None, [local], remote, single=True)
+
+
+    def _putMultipleFiles(self, files, remote):
+        """
+        Perform an upload for a list of local files.
+
+        @param files: List of local files.
+        @type files: C{list} of C{str}.
+
+        @param remote: Remote path for the request relative to current working
+            directory.
+        @type remote: C{str}
+
+        @return: A deferred which fires when transfer is done.
+        """
+        return self._cbPutMultipleNext(None, files, remote)
+
+
+    def _cbPutMultipleNext(
+            self, previousResult, files, remotePath, single=False):
+        """
+        Perform an upload for the next file in the list of local files.
+
+        @param previousResult: Result form previous file form the list.
+        @type previousResult: C{str}
+
+        @param files: List of local files.
+        @type files: C{list} of C{str}
+
+        @param remotePath: Remote path for the request relative to current
+            working directory.
+        @type remotePath: C{str}
+
+        @param single: A flag which signals if this is a transfer for a single
+            file in which case we use the exact remote path
+        @type single: C{bool}
+
+        @return: A deferred which fires when transfer is done.
+        """
+        if isinstance(previousResult, failure.Failure):
+            self._printFailure(previousResult)
+        elif previousResult:
+            self.transport.write(previousResult)
+            if not previousResult.endswith('\n'):
                 self.transport.write('\n')
-        f = None
-        while files and not f:
+
+        currentFile = None
+        while files and not currentFile:
             try:
-                f = files.pop(0)
-                lf = file(f, 'r')
+                currentFile = files.pop(0)
+                localStream = open(currentFile, 'r')
             except:
                 self._printFailure(failure.Failure())
-                f = None
-        if not f:
-            return
-        name = os.path.split(f)[1]
-        remote = os.path.join(self.currentDirectory, path, name)
-        log.msg((name, remote, path))
-        flags = filetransfer.FXF_WRITE|filetransfer.FXF_CREAT|filetransfer.FXF_TRUNC
-        d = self.client.openFile(remote, flags, {})
-        d.addCallback(self._cbPutOpenFile, lf)
-        d.addErrback(self._ebCloseLf, lf)
-        d.addBoth(self._cbPutMultipleNext, files, path)
+                currentFile = None
+
+        # No more files to transfer.
+        if not currentFile:
+            return None
+
+        if single:
+            remote = remotePath
+        else:
+            name = os.path.split(currentFile)[1]
+            remote = os.path.join(remotePath, name)
+            log.msg((name, remote, remotePath))
+
+        d = self._putRemoteFile(localStream, remote)
+        d.addBoth(self._cbPutMultipleNext, files, remotePath)
         return d
 
+
+    def _putRemoteFile(self, localStream, remotePath):
+        """
+        Do an upload request.
+
+        @param localStream: Local stream from where data is read.
+        @type localStream: File like object.
+
+        @param remotePath: Remote path for the request relative to current working directory.
+        @type remotePath: C{str}
+
+        @return: A deferred which fires when transfer is done.
+        """
+        remote = os.path.join(self.currentDirectory, remotePath)
+        flags = (
+            filetransfer.FXF_WRITE |
+            filetransfer.FXF_CREAT |
+            filetransfer.FXF_TRUNC
+            )
+        d = self.client.openFile(remote, flags, {})
+        d.addCallback(self._cbPutOpenFile, localStream)
+        d.addErrback(self._ebCloseLf, localStream)
+        return d
+
+
     def _cbPutOpenFile(self, rf, lf):
         numRequests = self.client.transport.conn.options['requests']
         if self.useProgressBar:
@@ -679,11 +764,11 @@ version                         Print the SFTP version.
     def _abbrevSize(self, size):
         # from http://mail.python.org/pipermail/python-list/1999-December/018395.html
         _abbrevs = [
-            (1<<50L, 'PB'),
-            (1<<40L, 'TB'),
-            (1<<30L, 'GB'),
-            (1<<20L, 'MB'),
-            (1<<10L, 'kB'),
+            (1<<50, 'PB'),
+            (1<<40, 'TB'),
+            (1<<30, 'GB'),
+            (1<<20, 'MB'),
+            (1<<10, 'kB'),
             (1, 'B')
             ]
 
@@ -734,7 +819,11 @@ version                         Print the SFTP version.
         else:
             timeLeft = 0
         front = f.name
-        back = '%3i%% %s %sps %s ' % ((total / f.size) * 100,
+        if f.size:
+            percentage = (total / f.size) * 100
+        else:
+            percentage = 100
+        back = '%3i%% %s %sps %s ' % (percentage,
                                       self._abbrevSize(total),
                                       self._abbrevSize(speed),
                                       self._abbrevTime(timeLeft))
@@ -743,9 +832,19 @@ version                         Print the SFTP version.
 
 
     def _getFilename(self, line):
-        line.lstrip()
+        """
+        Parse line received as command line input and return first filename
+        together with the remaining line.
+
+        @param line: Arguments received from command line input.
+        @type line: C{str}
+
+        @return: Tupple with filename and rest. Return empty values when no path was not found.
+        @rtype: C{tupple}
+        """
+        line = line.strip()
         if not line:
-            return None, ''
+            return '', ''
         if line[0] in '\'"':
             ret = []
             line = list(line)
@@ -757,17 +856,17 @@ version                         Print the SFTP version.
                     elif c == '\\': # quoted character
                         del line[i]
                         if line[i] not in '\'"\\':
-                            raise IndexError, "bad quote: \\%s" % line[i]
+                            raise IndexError("bad quote: \\%s" % (line[i],))
                         ret.append(line[i])
                     else:
                         ret.append(line[i])
             except IndexError:
-                raise IndexError, "unterminated quote"
+                raise IndexError("unterminated quote")
         ret = line.split(None, 1)
         if len(ret) == 1:
             return ret[0], ''
         else:
-            return ret
+            return ret[0], ret[1]
 
 StdioClient.__dict__['cmd_?'] = StdioClient.cmd_HELP
 
@@ -809,7 +908,7 @@ class SSHSession(channel.SSHChannel):
 
     def eofReceived(self):
         log.msg('got eof')
-        self.stdio.closeStdin()
+        self.stdio.loseWriteConnection()
 
     def closeReceived(self):
         log.msg('remote side closed %s' % self)
